use sqlx::Sqlite;
use stanza::client::presence::String1024;

#[derive(Debug, Default, sqlx::FromRow, Clone)]
pub struct Online {
    pub show: Option<Show>,
    #[sqlx(rename = "message")]
    pub status: Option<String>,
    #[sqlx(skip)]
    pub priority: Option<i8>,
}

#[derive(Debug, Clone, Copy)]
pub enum Show {
    Away,
    Chat,
    DoNotDisturb,
    ExtendedAway,
}

impl sqlx::Type<Sqlite> for Show {
    fn type_info() -> <Sqlite as sqlx::Database>::TypeInfo {
        <&str as sqlx::Type<Sqlite>>::type_info()
    }
}

impl sqlx::Decode<'_, Sqlite> for Show {
    fn decode(
        value: <Sqlite as sqlx::Database>::ValueRef<'_>,
    ) -> Result<Self, sqlx::error::BoxDynError> {
        let value = <&str as sqlx::Decode<Sqlite>>::decode(value)?;
        match value {
            "away" => Ok(Self::Away),
            "chat" => Ok(Self::Chat),
            "do-not-disturb" => Ok(Self::DoNotDisturb),
            "extended-away" => Ok(Self::ExtendedAway),
            _ => unreachable!(),
        }
    }
}

impl sqlx::Encode<'_, Sqlite> for Show {
    fn encode_by_ref(
        &self,
        buf: &mut <Sqlite as sqlx::Database>::ArgumentBuffer<'_>,
    ) -> Result<sqlx::encode::IsNull, sqlx::error::BoxDynError> {
        let value = match self {
            Show::Away => "away",
            Show::Chat => "chat",
            Show::DoNotDisturb => "do-not-disturb",
            Show::ExtendedAway => "extended-away",
        };
        <&str as sqlx::Encode<Sqlite>>::encode(value, buf)
    }
}

#[derive(Debug, Default, Clone)]
pub struct Offline {
    pub status: Option<String>,
}

#[derive(Debug, Clone)]
pub enum Presence {
    Online(Online),
    Offline(Offline),
}

impl From<Online> for stanza::client::presence::Presence {
    fn from(value: Online) -> Self {
        Self {
            from: None,
            id: None,
            to: None,
            r#type: None,
            lang: None,
            show: value.show.map(|show| match show {
                Show::Away => stanza::client::presence::Show::Away,
                Show::Chat => stanza::client::presence::Show::Chat,
                Show::DoNotDisturb => stanza::client::presence::Show::Dnd,
                Show::ExtendedAway => stanza::client::presence::Show::Xa,
            }),
            // TODO: enforce message length in status message
            status: value.status.map(|status| stanza::client::presence::Status {
                lang: None,
                status: String1024(status),
            }),
            priority: value
                .priority
                .map(|priority| stanza::client::presence::Priority(priority)),
            errors: Vec::new(),
        }
    }
}

impl From<Offline> for stanza::client::presence::Presence {
    fn from(value: Offline) -> Self {
        Self {
            from: None,
            id: None,
            to: None,
            r#type: Some(stanza::client::presence::PresenceType::Unavailable),
            lang: None,
            show: None,
            status: value.status.map(|status| stanza::client::presence::Status {
                lang: None,
                status: String1024(status),
            }),
            priority: None,
            errors: Vec::new(),
        }
    }
}
